//Copyright 1998 President and Fellows of Harvard University
//motif.h

#ifndef _motif
#define _motif
#include "sites.h"
#include "myheap.h"
#include "genome.h"

class Motif : public Sites{
protected:
  double *mat_fast;
  double motif_gc;
 public:
  Motif(vector<string> s, int w=10, bool both=true): Sites(s,w,both) {
    mat_fast=new double[4*w];motif_gc=0.5;}
  ~Motif(){delete[] mat_fast;}
  Motif(const Motif&);
  Motif& operator= (const Motif&);
  static void usage_error();
  void put_motif(ostream &fout);
  double score(const string &s);
  double score_raw(const string &s);
  double score_partial(const string &site, int offset, int &num_cols_used);
  double score(const Site &s){return score(seq(s));}
  double score(const short *s);
  double score_rev(const short *s);
  void update() {updated=false;create_matrix();}
  virtual void create_matrix();//allows the scoring method to be changed in derived classes
  void freq_matrix_fast(double *fm, const bool useactive=true, const int five=0, const int three=0);
  void stats_all(double &avg, double &sd);//for all of seqset
  void stats_self(double &avg, double &sd);
  void set_gc(double g){motif_gc=g;}
};

class MatrixTest :public StringTest{
  Motif mot;
  double cutoff;
  double sum;
  double sum_squares;
  int num_calls;
 public:
  MatrixTest(Motif m, double x):mot(m),cutoff(x),sum(0.0),
    sum_squares(0.0),num_calls(0){}
  bool operator()(string s){return (mot.score(s)>cutoff);}
  bool operator()(const short *s,bool t);
  double avg_scores(){return sum/num_calls;}
  double sd_scores(){
    return (sqrt(sum_squares/(num_calls-1)-(sum/num_calls)*(sum/(num_calls-1)) ));
  }
  void reset(){sum=0.0;sum_squares=0.0;num_calls=0;}
  void set_cutoff(double x){cutoff=x;}
};

void put_score_sites(Genome &gn,Motif &m,ostream &fout, int cut, int max, int top_sites, bool locinfo);
void score_sites(int argc, char* argv[]);
Motif read_alignace_motif(istream &fin, int num);
	
#endif
