//Copyright 1998 President and Fellows of Harvard University
//orfs.cpp

#include "orfs.h"

bool operator< (const Orf s, const Orf t){
  if(s.seqnum()<t.seqnum()) return true;
  if(s.seqnum()>t.seqnum()) return false;
  if(s.left()<t.left()) return true;
  if(s.left()>t.left()) return false;
  if(s.right()<t.right()) return true;
  return false;
}


bool operator== (const Orf s, const Orf t){
  if(s.seqnum()!=t.seqnum()) return false;
  if(s.left()!=t.left()) return false;
  if(s.right()!=t.right()) return false;
  //	if(s.strand()!=t.strand()) return false;
  //	if(s.is_orf()!=t.is_orf()) return false;
  return true;
}

bool operator!= (const Orf s, const Orf t){
  return !(s==t);
}
bool operator> (const Orf s, const Orf t){
  if(s==t) return false;
  return !(s<t);
}

bool nested_orfs(const Orf &o, const Orf &p){
  int r=o.right()-p.right();
  int l=o.left()-p.left();
  if((r*l)<=0) return true;
  //if either endpoint match then the orfs are nested
  //if the relationships r and l are reversed then rl<0 and they are nested
  return false;
}

bool overlapped_orfs(const Orf &o, const Orf &p){
  int d1=o.right()-p.left();
  int d2=p.right()-o.left();
  if(d1>=0 && d2>=0) return true;
  return false;
}

int overlap_size(const Orf &o, const Orf &p){
  if(!overlapped_orfs(o,p)) return 0;
  if(nested_orfs(o,p)){
    if(o.length()<p.length()) return o.length();
    return p.length();
  }
  int d1=o.right()-p.left()+1;
  int d2=p.right()-o.left()+1;
  if(d1<d2) return d1;
  return d2;
}



Orfs::Orfs(const Orfs& s){
  //copy constructor  Orfs a=b;
  num_sequences=s.num_chroms();
  orf_list=new list<Orf>[num_sequences+1];
  for(int i=1;i<=num_sequences;i++){
    orf_list[i]=s.orfs_chrom_const(i);
  }
  mapped=false;
}

Orfs& Orfs::operator= (const Orfs& s){
  //copy assignment  a=b  (already initialized once)
  if(this!=&s){
    delete[] orf_list;
    num_sequences=s.num_chroms();
    orf_list=new list<Orf>[num_sequences];
    for(int i=1;i<=num_sequences;i++){
      orf_list[i]=s.orfs_chrom_const(i);
    }
    mapped=false;
  }
  return *this;
}

void Orfs::update_map(){
  if(mapped) return;
  mapped=true;
  for(int i=1;i<=num_sequences;i++){
    list<Orf>::const_iterator o=orf_list[i].begin();
    for(;o!=orf_list[i].end();o++){
      string s=(*o).yname();
      string t=(*o).genename();
      //make all caps
      s=capitalize(s);
      t=capitalize(t);
      lookup[s]=(*o);
      lookup[t]=(*o);
    }
  }
}

Orf Orfs::find_orf(string yn){
  if(!mapped) update_map();
  //make all caps
  string::iterator s=yn.begin();
  for(;s!=yn.end();s++){
    *s=toupper(*s);
  }
  return lookup[yn];
}


void Orfs::sort_orfs(){
  if(sorted) return;
  sorted=true;
  for(int i=1;i<=num_sequences;i++){
    orf_list[i].sort();
  }
}

OrfLocale::OrfLocale(int s, int n, Orfs &o){
  init(s,n,n,o);
}

OrfLocale::OrfLocale(Orf f, Orfs &o):this_one(f){
  init(f.seqnum(),f.left(),f.right(),o);
}

OrfLocale::OrfLocale(Site st,int wide, Orfs &o):oloc_site(st){
  init(st.seqnum(),st.pos(),st.pos()+wide-1,o);
}


void OrfLocale::init(const int s,const int l, const int r, Orfs &o){
  //if such orfs do not exist, the null designations will remain,
  //test with is_null()
  //use add_ends() to ensure that all such members are available
  chrom=s;
  pos_l=l;
  pos_r=r;
  list<Orf> lo=o.orfs_chrom(s);//already sorted
  list<Orf>::iterator a=lo.begin();
  int closest_left_orf=INT_MAX;
  int closest_right_orf=INT_MAX;
  int closest_left_other=INT_MAX;
  int closest_right_other=INT_MAX;
  int closest_left_orf_c=INT_MAX;
  int closest_right_orf_w=INT_MAX;
  int d;
  for(;a!=lo.end();a++){
    if((*a).is_orf()){
      if(l>(*a).right()){//orf is to the left of f
	d=l-(*a).right();
	if(d<closest_left_orf){
	  closest_left_orf=d;
	  near_left_orf=(*a);
	}
	if(!(*a).strand()&&d<closest_left_orf_c){
	  closest_left_orf_c=d;
	  near_left_orf_c=(*a);
	}
      }
      else if(r<(*a).left()){//orf is to the right of f
	d=(*a).left()-r;
	if(d<closest_right_orf){
	  closest_right_orf=d;
	  near_right_orf=(*a);
	}
	if((*a).strand()&&d<closest_right_orf_w){
	  closest_right_orf_w=d;
	  near_right_orf_w=(*a);
	}
      }
      else{//orf overlaps f
	in_orf.push_back(*a);
      }
    }
    else{//non-orf
      if(l>(*a).right()){//orf is to the left of f
	d=l-(*a).right();
	if(d<closest_left_other){
	  closest_left_other=d;
	  near_left_other=(*a);
	}
      }
      else if(r<(*a).left()){//orf is to the right of f
	d=(*a).left()-r;
	if(d<closest_right_other){
	  closest_right_other=d;
	  near_right_other=(*a);
	}
      }
      else{//orf overlaps f
	in_other.push_back(*a);
      }
    }
  }
  //This should initialize all data members if there are enough features
  //in the orf_list to do so.
  //Definitely will if add_ends() is called first.
}



void Orfs::load_table(istream &fout, bool tabletype){
  //format from SGD, works with files from April 98...
  string s=string("start");
  while(!(s[0]=='='&&s[1]=='='&&s[2]=='=')){
    getline(fout,s,'\n');
  }
  //table begins
  while(getline(fout,s,'\n')){
    vector<string> col=split(s,'\t');
    int c=convert_roman(col[3]);
    int b=str_to_int(col[4]);
    int e=str_to_int(col[5]);
    bool st=true;
    if(b>e){//switch
      int temp=e;
      e=b;b=temp;
      st=false;
    }
    Orf o(c,b,e,st);
    if(!tabletype) o.set_nonorf();
    o.set_name1(col[0]);
    o.set_name2(col[2]);
    if(col.size()>7) o.set_desc1(col[7]);
    if(col.size()>8) o.set_desc2(col[8]);
    orf_list[c].push_back(o);
  }
}

int Orfs::free_up(Orf o){
  //returns the amount of free space upstream of an Orf, i.e., the amount
  //of sequence before you run into an orf or other feature
  //ignore overlapping features!!!
  int pos1=o.start();
  //	if(o.strand()) pos1--;
  //	else pos1++;//shift pos1 to just before beginning of orf
  //	int chr=o.seqnum();
  int ret;
  OrfLocale oloc(o,*this);
  //	if(oloc.num_inside_features()!=0) return 0;//inside a feature
  if(o.strand()){//look left
    ret=pos1-(oloc.left_orf()).right()-1;
    int ret2=pos1-(oloc.left_other()).right()-1;
    if(ret2<ret) ret=ret2;
  }
  else{//look right
    ret=(oloc.right_orf()).left()-pos1-1;
    int ret2=(oloc.right_other()).left()-pos1-1;
    if(ret2<ret) ret=ret2;
  }
  return ret;
  //non-negative number
}

int Orfs::free_down(Orf o){
  //returns the amount of free space upstream of an Orf, i.e., the amount
  //of sequence before you run into an orf or other feature
  //ignore overlapping features!!!
  int pos1=o.start();
  //	if(o.strand()) pos1--;
  //	else pos1++;//shift pos1 to just before beginning of orf
  //	int chr=o.seqnum();
  int ret;
  OrfLocale oloc(o,*this);
  //	if(oloc.num_inside_features()!=0) return 0;//inside a feature
  if(!o.strand()){//look right
    ret=pos1-(oloc.left_orf()).right()-1;
    int ret2=pos1-(oloc.left_other()).right()-1;
    if(ret2<ret) ret=ret2;
  }
  else{//look left
    ret=(oloc.right_orf()).left()-pos1-1;
    int ret2=(oloc.right_other()).left()-pos1-1;
    if(ret2<ret) ret=ret2;
  }
  return ret;
  //non-negative number
}

void Orfs::add_ends(const vector<string> &ss){
  //adds pseudo orfs for the beginning and endings of each sequence in ss
  //both orf and other types are added, pointing away from sequence
  vector<string>::const_iterator v=ss.begin();
  char ch='A';
  for(int i=1;v!=ss.end();v++,i++,ch++){
    Orf o(i,0,0,false);
    string nm1=string("Y")+ch+string("begin");
    o.set_name1(nm1);
    o.set_desc2(string("end"));
    orf_list[i].push_back(o);
    o.set_nonorf();
    orf_list[i].push_back(o);
    
    Orf p(i,(*v).size()+1,(*v).size()+1,true);
    string nm2=string("Y")+ch+string("end");
    p.set_name1(nm2);
    p.set_desc2(string("end"));
    orf_list[i].push_back(p);
    p.set_nonorf();
    orf_list[i].push_back(p);
  }
}

string OrfLocale::desc(){
  string s;
  int i;
  if(in_orf.size()!=0||in_other.size()!=0){
    s="in ";
    for(i=0;i<in_orf.size();i++){
      s+=in_orf[i].yname()+" ";
    }
    for(i=0;i<in_other.size();i++){
      s+=in_other[i].yname()+" ";
    }
  }
  else{
    Orf p1=left_feat();
    if(p1.strand()) s+="downstream of ";
    else s+="upstream of ";
    s+=p1.yname()+"  ";
    Orf p2=right_feat();
    if(p2.strand()) s+="upstream of ";
    else s+="downstream of ";
    s+=p2.yname();
  }
  return s;
}

string OrfLocale::desc_long(){
  string s;
  int i;
  
  for(i=0;i<in_orf.size();i++){
    s+="inside "+in_orf[i].full_desc()+"\n";
  }
  for(i=0;i<in_other.size();i++){
    s+="inside "+in_other[i].full_desc()+"\n";
  }
  
  Orf p1=left_feat();
  s+=int_to_str(dist_left_feat())+" bp ";
  if(p1.strand()) s+="downstream of ";
  else s+="upstream of ";
  s+=p1.full_desc()+"\n";
  
  Orf p2=right_feat();
  s+=int_to_str(dist_right_feat())+" bp ";
  if(p2.strand()) s+="upstream of ";
  else s+="downstream of ";
  s+=p2.full_desc()+"\n";
  
  return s;
}

void Orfs::load_funcats(istream &fin){
  string s;
  while(getline(fin,s,'\n')){
    vector<string> vs=split(s,'|');
    for(int i=0;i<vs.size();i++){
      vs[i]=clip_white(vs[i]);
      vs[i]=lower_case(vs[i]);
    }
    Orf o=find_orf(vs[0]);
    if(o.is_null()){
      cerr<<"Orfs::load_funcats() -> no such orf as "<<vs[0];
      cerr<<"  info not loaded\n";
      continue;
      //			exit(0);
    }
    list<Orf>::iterator p=find(orf_list[o.seqnum()].begin(),
			       orf_list[o.seqnum()].end(),o);
    if(vs.size()>2&&vs[2].size()>0) (*p).add_funcat(vs[2]);
  }
}

bool Orf::in_funcat(string s){
  s=clip_white(s);
  s=lower_case(s);
  for(int i=0;i<category.size();i++){
    if(s==category[i]) return true;
  }
  return false;
}


