//Copyright 1998 President and Fellows of Harvard University
//sites.cpp

#include "sites.h"

void Sites::add_sites_better(StringTest &test, bool oneperstrand){
  for(int i=1;i<=num_seqs();i++){
    for(int j=1;j<=seq_len(i)-width+1;j++){
      if(test(&(seq_fast[i][j]),true)){
	add(Site(i,j,true));
	if(oneperstrand) continue;
      }
      if(double_strand){
	if(test(&(seq_fast[i][j]),false)){
	  add(Site(i,j,false));
	}
      }
    }
  }
}


string Sites::seq(const int chrom, const int pos, const bool strand,
				  const bool useactive, const int five, const int three){
	string ret;
	vector<string>::iterator s=seqset.begin()+chrom-1;
//	string::iterator s=(*first+chrom-1).begin()+pos-1;
	list<bool>::const_iterator a=active.begin();
	list<bool>::reverse_iterator b=active.rbegin();
	bool inside=false;
	int x5,x3;
	if(strand){x5=five;x3=three;}
	else{x5=three;x3=five;}//switch three and five
	for(int i=pos-1-x5;i!=pos-1+width+x3;i++){
		if(i==(pos-1)) inside=true;
		if(i==(pos-1+width)) inside=false;
		if(i<0||i>=seq_len(chrom)) ret+=' ';
		else if(!useactive||!inside) ret+=(*s)[i];
		else if((strand&&*a)||(!strand&&*b)) ret+=(*s)[i];
		//reads active backwards if !strand
		if(inside) {a++;b++;}
	}
	if(!strand) ret=reverse_comp(ret);
	return ret;
}

string Sites::seq(Site s,const bool useactive,const int five, const int three){
	string ret;
	ret=seq(s.seqnum(),s.pos(),s.strand(),useactive,five, three);
	return ret;
}


bool operator< (const Site s,const Site t){
	if (s.seqnum() < t.seqnum()) return true;
	if (s.seqnum() > t.seqnum()) return false;
	if (s.pos() < t.pos()) return true;
	if (s.pos() > t.pos()) return false;
//	if (!s.watson && t.watson) return true;
	return false;
}

bool operator== (const Site s,const Site t){
	if (s.seqnum() != t.seqnum()) return false;
	if (s.pos() != t.pos()) return false;
//	if ((*this).sequence != t.sequence) return false;
//	if ((*this).watson != t.watson) return false;
	return true;
}

bool operator!= (const Site s, const Site t){
	return (!(s==t));
}

bool operator> (const Site s, const Site t){
	return (!(s<t));
}


bool Sites::belongs(Site s){
	list<Site>::iterator p;
	for(p=sites.begin();p!=sites.end();p++){
		if(s==(*p)) return true;
	}
	return false;
}

int Sites::tot_avail(){
	vector<string>::iterator p;
	int res=0;
	for(p=seqset.begin();p!=seqset.end();p++){
		res+=( (*p).size() -width +1 );
	}
	return res;
}


void Sites::expand(int five, int three){
	//expansion should be checked before calling this fxn
	//can also be used to contract
	updated=true;
	width+=(five+three);
	int i;
	for(i=0;i<five;i++) active.push_front(false);
	for(i=0;i<three;i++) active.push_back(false);
	for(i=0;i>five;i--) {
		if(active.front()){
			cout<<"trouble 5"<<'\n';
		}
		active.pop_front();
	}
	for(i=0;i>three;i--) {
		if(active.back()) {
			cout<<"trouble 3"<<'\n';
		}
		active.pop_back();
	}
	list<Site>::iterator s=sites.begin();
	for(;s!=sites.end();s++){
		if((*s).strand()) (*s).shift_left(five);
		else (*s).shift_left(three);
	}
}


void Sites::clear_all(){
	updated=true;
	width=num_cols;
	num_sites=0;
	//double_strand unchanged
	//num_cols presumably remains unchanged through all Sites manipulations
	sites.clear();
//	active.assign(num_cols,true);
	active.clear();
	for(int i=0;i<num_cols;i++) active.push_back(true);
	//abby
}

void Sites::clear_sites(){
	//same as clear_all, except the width and column configuration is preserved
	updated=true;
	num_sites=0;
	sites.clear();
}
	

void Sites::create_intarrays(){
  map<char,int> convert;
  convert['a']=convert['A']=0;
  convert['c']=convert['C']=1;
  convert['g']=convert['G']=2;
  convert['t']=convert['T']=3;
  convert['n']=convert['N']=-1;
  vector<string>::iterator s=seqset.begin();
  string::iterator t;
  seq_fast=new short*[seqset.size()+1];
  for(int i=1;s!=seqset.end();s++,i++){
    seq_fast[i]=new short[(*s).size()+1];
    t=(*s).begin();
    for(int j=1;t!=(*s).end();t++,j++){
      seq_fast[i][j]=convert[(*t)];
    }
  }
}


Sites::~Sites(){
	for(int i=1;i<=seqset.size();i++){
		delete[] seq_fast[i];
	}
	delete[] seq_fast;
}


Sites::Sites(const Sites& s){
	//copy constructor  Sites a=b;
	width=s.width;
	num_cols=s.num_cols;
	double_strand=s.double_strand;
	num_sites=s.num_sites;
	sites=s.sites;
	seqset=s.seqset;
	active=s.active;
	updated=s.updated;
	create_intarrays();
}

Sites& Sites::operator= (const Sites& s){
	//copy assignment  a=b  (already initialized once)
	if(this!=&s){
		for(int i=1;i<=seqset.size();i++){
			delete[] seq_fast[i];
		}
		delete[] seq_fast;
		width=s.width;
		num_cols=s.num_cols;
		double_strand=s.double_strand;
		num_sites=s.num_sites;
		sites=s.sites;
		seqset=s.seqset;
		active=s.active;
		updated=s.updated;
		create_intarrays();
	}
	return *this;
}

void Sites::set_columns(const string &s){
	//reads in a string containing spaces and asterisks, translates into
	//column structure and sets proper width
	active.clear();
	width=s.size();
	num_cols=0;
	for(int i=0;i<s.size();i++){
		if(s[i]=='*') {
			active.push_back(true);
			num_cols++;
		}
		else if(s[i]==' ') active.push_back(false);
		else {
			cerr<<"invalid input to Sites::set_columns()\n";
			cerr<<"input:  "<<s;
			exit(0);
		}
	}
}

void Sites::trim_column_stx(){
	int x3,x5;
	list<bool>::iterator a;
	for(x5=0,a=active.begin();!(*a);a++){x5--;}
	list<bool>::reverse_iterator b;
	for(x3=0,b=active.rbegin();!(*b);b++){x3--;}
	expand(x5,x3);
}


vector<string> Sites::raw_sites(int five, int three){
//output actual sites, no consideration of active columns
	vector<string> vs;
	list<Site>::iterator ls=sites.begin();
	for(;ls!=sites.end();ls++){
		vs.push_back(seq(*ls,false,five,three));
	}
	return vs;
}


string Sites::keyname(int i){
  string s=nameset[i-1];
  vector<string> a=split(s,' ');
  return a[0];
}
