//Copyright 1998 President and Fellows of Harvard University
//sites.h

#ifndef _sites
#define _sites
#include "standard.h"

class Site{
	int chrom;
	int posit;
	bool watson;
public:
	Site(int a, int b, bool c):chrom(a),posit(b),watson(c){}
	Site():chrom(1),posit(1),watson(true){};
	int seqnum() const {return chrom;}
	int pos() const {return posit;}
	bool strand() const {return watson;}
	void shift_left(int k=1){posit-=k;}
	void shift_right(int k=1){posit+=k;}
	void next_seq(){chrom++;}
	void prev_seq(){chrom--;}
	void set_pos(int k){posit=k;}
	void switch_strand(){watson=!watson;}
	Site other_strand() const {return Site(chrom,posit,!watson);}
};

bool operator< (const Site s, const Site t);
bool operator== (const Site s, const Site t);
// Equality for sites depends on seqnum and position, not strand or sequence.
bool operator!= (const Site s, const Site t);
bool operator> (const Site s, const Site t);

class StringTest{
public:
	virtual bool operator()(string s)=0;
	virtual bool operator()(const short *s,bool t)=0;
};

class Motif;

class Sites{
	int width;
	int num_cols;
	bool double_strand;
protected:
	int num_sites;
	list<Site> sites;
	vector<string> seqset;
	vector<string> nameset;
	short **seq_fast;
	list<bool> active;
	bool updated;//set to true whenever the list of sites is changed
	//derived classes reset it to false to minimize expensive calculations
	//that depend on the site list
public:
	Sites(vector<string> s, int c=10, bool both=true):seqset(s),num_cols(c),
	  width(c),num_sites(0),double_strand(both) {
	  for(int i=0;i<s.size();i++) nameset.push_back(int_to_str(i+1));
	  for(int i=0;i<c;i++) active.push_back(true);
	  updated=false;create_intarrays();}
	Sites(vector<string> s, vector<string> n,int c=10, bool both=true):seqset(s),nameset(n),num_cols(c),
	  width(c),num_sites(0),double_strand(both) {
	  for(int i=0;i<c;i++) active.push_back(true);
	  updated=false;create_intarrays();}
//	Sites(){}//for copy constructor
	~Sites();
	Sites(const Sites&);
	Sites& operator= (const Sites&);
	void copy_column_stx(Sites &s){width=s.wide();num_cols=s.ncols();active=s.active_cols();}
	void create_intarrays();
	void clear_all();
	void clear_sites();
	int number() const {return num_sites;}
	int wide() const {return width;}
	int ncols() const {return num_cols;}
	list<Site> where() const {return (sites);}
	list<bool> active_cols() const {return active;}
	int num_seqs(){return seqset.size();}
	int seq_len(int k){vector<string>::iterator p=seqset.begin();p+=(k-1);return((*p).size());}
	int tot_avail();//returns the total number of positions possible on one strand
	//given the size of the seqset and the width of the sites
	bool belongs(Site s);
	void expand(int five, int three);//pads active with falses
	//negative numbers ok, contract regardless of activity of columns
	string seq(Site s,const bool useactive=true, const int five=0, const int three=0);
	string seq(const int chrom,const int pos,const bool strand,
		const bool useactive=true, const int five=0, const int three=0);
	void add(Site s) {updated=true;sites.push_back(s);num_sites++;}
	//this add is unconditional.  Be sure that the site being added is
	//not already there, else num_sites gets out of whack.
	void add_sites_better(StringTest &st, bool oneperstrand=false);
	void del(Site s) {updated=true;sites.remove(s); num_sites--;}
	//this del is unconditional.  Be sure that the site being deleted is
	//an existing site, else num_sites gets out of whack.
	string active_string(){string s;for(list<bool>::iterator a=active.begin();a!=active.end();a++) {if(*a) s+='*';else s+=' ';}return s;}
	const short* quick_seq(int i, int j){return &(seq_fast[i][j]);}
	void set_columns(const string &s);
	void trim_column_stx();
	vector<string> raw_sites(int five=0, int three=0);
	string keyname(int i);
};


#endif
