//Copyright 1998 President and Fellows of Harvard University
//genome.cpp

#include "genome.h"

string Genome::getup(string yn, int offset1, int offset2){
  Orf o=find_orf(yn);
  int pos1,len;
  len=offset2-offset1+1;
  if(len<=0){
    cerr<<"bad parameters for Genome::getup, offset1>offset2\n";
    return string("");
  }
  if(o.strand()) pos1=o.left()+offset1;
  else pos1=o.right()-offset2;
  if(pos1<1){
    cerr<<"bad parameters for Genome::getup, outside seqset "<<pos1<<"\n";
    return string("");
  }
  string ret=seqset[o.seqnum()-1].substr(pos1-1,len);
  //this changed 021700 so that sequences extend from start
  if(o.strand()){
    ret=reverse_comp(ret);
  }
  return ret;
}

string Genome::getdown(string yn, int offset1, int offset2){
  Orf o=find_orf(yn);
  int pos1,len;
  len=offset2-offset1+1;
  if(len<=0){
    cerr<<"bad parameters for Genome::getup, offset1>offset2\n";
    return string("");
  }
  if(!o.strand()) pos1=o.left()+offset1;//this is the only line changed from getup
  else pos1=o.right()-offset2;
  if(pos1<1){
    cerr<<"bad parameters for Genome::getup, outside seqset "<<pos1<<"\n";
    return string("");
  }
  string ret=seqset[o.seqnum()-1].substr(pos1-1,len);
  if(!o.strand()){
    ret=reverse_comp(ret);
  }
  return ret;
}

string Genome::getup(Orf o, int offset1, int offset2){
  string yn=o.yname();
  return getup(yn,offset1,offset2);
}

string Genome::getdown(Orf o, int offset1, int offset2){
  string yn=o.yname();
  return getdown(yn,offset1,offset2);
}

string Genome::getup_nonfeat(string yn, int min, int max){
  //min and max refer to the min and max amount of sequence to be returned
  Orf o=find_orf(yn);
  int openspace=free_up(o);
  int off1=openspace;
  if(off1<min) off1=min;
  if(off1>max) off1=max;
  //check for edge of seqset
  int sp;
  if(o.strand()) sp=o.left()-1;
  else sp=seq_len(o.seqnum())-o.right();
  if(off1>sp) off1=sp;
  return getup(yn,-off1);
}

string Genome::getdown_nonfeat(string yn, int min, int max){
  //min and max refer to the min and max amount of sequence to be returned
  Orf o=find_orf(yn);
  int openspace=free_down(o);//changed
  int off1=openspace;
  if(off1<min) off1=min;
  if(off1>max) off1=max;
  //check for edge of seqset
  int sp;
  if(!o.strand()) sp=o.left()-1;//changed
  else sp=seq_len(o.seqnum())-o.right();
  if(off1>sp) off1=sp;
  return getup(yn,-off1);
}

string Genome::getup_nonfeat(Orf o, int min, int max){
  string yn=o.yname();
  return getup_nonfeat(yn,min,max);
}

string Genome::getdown_nonfeat(Orf o, int min, int max){
  string yn=o.yname();
  return getdown_nonfeat(yn,min,max);
}

string Genome::getseq(string yn){
  Orf o=find_orf(yn);
  if(o.is_null()) {
    cout<<"Problem with Genome::getseq\n";
    exit(0);
  }
  int len=o.right()-o.left()+1;
  int pos1=o.left();
  string ret=seqset[o.seqnum()-1].substr(pos1-1,len);
  if(!o.strand()){
    ret=reverse_comp(ret);
  }
  return ret;
}

string Genome::getseq(Orf o){
  string yn=o.yname();
  return getseq(yn);
}

void Genome::pickup(istream &fin,ostream &fout, int min, int max){
  vector<string> temp;
  pickup(fin,fout,temp,min,max);
}

void Genome::pickdown(istream &fin,ostream &fout, int min, int max){
  vector<string> temp;
  pickdown(fin,fout,temp,min,max);
}

void Genome::pickup(istream &fin,ostream &fout,vector<string> &notfound,
					int min, int max){
//does not duplicate intergenic regions
  string s;
  list<Orf> lo;
  while(getline(fin,s)){
    if(s.size()==0) continue;
    Orf q=find_orf(s);
    if(q.is_null()) {
      cerr<<"No orf named "<<s<<'\n';
      notfound.push_back(s);
    }
    else lo.push_back(find_orf(s));
  }
  lo.sort();
  lo.unique();
  list<Orf>::iterator p=lo.begin();
  Orf lst;
  string up;
  for(;p!=lo.end();p++){
    Orf cur=(*p);
    OrfLocale oloc(cur.seqnum(),cur.left(),*this);
    if(oloc.left_feat()==lst){
      //if overlap, put one unified region out
      if(!lst.strand()&&cur.strand()){//divergent
	int dist=cur.left()-lst.right()-1;
	if(dist<(2*max)){//unify
	  if(dist<min){
	    //get 300 for each, but exclude the overlap region
	    //includes the starts of two orfs, but one might be bogus
	    up=getup(cur,-min,min-dist-1);
	    fout<<'>'<<lst.yname()<<'_'<<cur.yname()<<" "<<
	      up.length()<<" bp "<<lst.full_desc()<<" "<<
	      cur.full_desc()<<'\n'<<up<<'\n';
	    lst.set_null();
	    continue;
	  }	
	  up=getup(lst,-dist);
	  fout<<'>'<<lst.yname()<<'_'<<cur.yname()<<" "<<
	    up.length()<<" bp "<<lst.full_desc()<<" "<<
	    cur.full_desc()<<'\n'<<up<<'\n';
	  lst.set_null();
	  continue;
	}
      }
    }
    if(!lst.is_null()){
      up=getup_nonfeat(lst,min,max);
      fout<<'>'<<lst.yname()<<" "<<
	up.length()<<" bp "<<lst.full_desc()<<'\n'<<up<<'\n';
    }
    lst=cur;
  }
  if(!lst.is_null()){
    up=getup_nonfeat(lst,min,max);
    fout<<'>'<<lst.yname()<<" "<<
      up.length()<<" bp "<<lst.full_desc()<<'\n'<<up<<'\n';
  }
}

void Genome::pickdown(istream &fin,ostream &fout,vector<string> &notfound,
					int min, int max){
  //does not duplicate intergenic regions
  string s;
  list<Orf> lo;
  while(getline(fin,s)){
    if(s.size()==0) continue;
    Orf q=find_orf(s);
    if(q.is_null()) {
      cerr<<"No orf named "<<s<<'\n';
      notfound.push_back(s);
    }
    else lo.push_back(find_orf(s));
  }
  lo.sort();
  lo.unique();
  list<Orf>::iterator p=lo.begin();
  Orf lst;
  string up;
  for(;p!=lo.end();p++){
    Orf cur=(*p);
    OrfLocale oloc(cur.seqnum(),cur.left(),*this);
    if(oloc.left_feat()==lst){
      //if overlap, put one unified region out
      if(lst.strand()&&!cur.strand()){//convergent
	int dist=cur.left()-lst.right()-1;
	if(dist<(2*max)){//unify
	  if(dist<min){
	    //get min for each, but exclude the overlap region
	    //includes the starts of two orfs, but one might be bogus
	    up=getdown(cur,-min,min-dist-1);
	    fout<<'>'<<lst.yname()<<'_'<<cur.yname()<<" "<<
	      up.length()<<" bp "<<lst.full_desc()<<' '<<
	      cur.full_desc()<<'\n'<<up<<'\n';
	    lst.set_null();
	    continue;
	  }	
	  up=getdown(cur,-dist);
	  fout<<'>'<<lst.yname()<<'_'<<cur.yname()<<" "<<
	    up.length()<<" bp "<<lst.full_desc()<<' '<<
	    cur.full_desc()<<'\n'<<up<<'\n';
	  lst.set_null();
	  continue;
	}
      }
    }
    if(!lst.is_null()){
      up=getdown_nonfeat(lst,min,max);
      fout<<'>'<<lst.yname()<<" "<<
	up.length()<<" bp "<<lst.full_desc()<<'\n'<<up<<'\n';
    }
    lst=cur;
  }
  if(!lst.is_null()){
    up=getdown_nonfeat(lst,min,max);
    fout<<'>'<<lst.yname()<<" "<<
      up.length()<<" bp "<<lst.full_desc()<<'\n'<<up<<'\n';
  }
}


void Genome::test_start_stop(ostream &fout){
  map<string,int> init;
  map<string,int> term;
  map<int,int> bin;
  for(int i=1;i<num_chroms();i++){
    list<Orf> lo=orfs_chrom(i);
    for(list<Orf>::iterator p=lo.begin();p!=lo.end();p++){
      if((*p).is_end()) continue;
      if(!(*p).is_orf()) continue;
      string s=getseq(*p);
      string si=s.substr(0,3);
      string st=s.substr(s.size()-3,3);
      init[si]++;
      term[st]++;
      fout<<(*p).description()<<'\n'<<s<<'\n';
      fout<<si<<'\t'<<st<<'\n';
      if(si!="ATG"){
	fout<<(*p).full_desc()<<'\n'<<s<<'\n'<<'\n';
      }
    }
  }
  fout<<"\n\n\n\n\n";
  fout<<"Start\n";
  map<string,int>::iterator ms=init.begin();
  for(;ms!=init.end();ms++){
    fout<<ms->first<<'\t'<<ms->second<<'\n';
  }
  fout<<"\n\nStop\n";
  for(ms=term.begin();ms!=term.end();ms++){
    fout<<ms->first<<'\t'<<ms->second<<'\n';
  }
}


void pickup(int argc, char* argv[]){
  string i,o;int k=300;int l=600;
  GetArg2(argc,argv,"-i",i);
  GetArg2(argc,argv,"-k",k);
  GetArg2(argc,argv,"-l",l);
  if(argc<2||i.size()==0){
    cerr<<"Returns upstream sequences in fasta format given a list of orfs as input.\n";
    cerr<<"Minimum and maximum sizes of input sequence may be set.  Otherwise, the \n";
    cerr<<"upstream read stops at the nearest other feature.\n";
    cerr<<"Usage: Pickup -i (input filename) -(options)\n";
    cerr<<"-k <int> min amount of upstream sequence to take (300)\n";
    cerr<<"-l <int> max amount of upstream sequence to take (600)\n";
    cerr<<"-a <string> orf table (ORF_table.txt)\n";
    cerr<<"-b <string> non-orf table (other_features_table.txt)\n";
    cerr<<"-z <string> genome sequence (yeast.seq)\n";
    exit(0);
  }
  
  string a,b,z;
  a="ORF_table.txt";
  b="other_features_table.txt";
  z="yeast.seq";
  GetArg2(argc,argv,"-a",a);//orf table
  GetArg2(argc,argv,"-b",b);//non-orf table
  GetArg2(argc,argv,"-z",z);//genome
  
  vector<string> genseq;
  get_fasta_fast(z.c_str(),genseq);
  Genome gen(genseq);
  if(a.size()!=0){
    ifstream tab_orf(a.c_str());
    gen.load_table(tab_orf,true);
  }
  if(b.size()!=0){
    ifstream tab_nonorf(b.c_str());
    gen.load_table(tab_nonorf,false);
  }
  gen.add_ends(genseq);
  
  ifstream fin(i.c_str());
  gen.pickup(fin,cout,k,l);
}


void find_seqs(int argc, char *argv[]){
  if(argc<2){
    cerr<<"findseqs -i (input file) -o (output file) -s (sequence)\n";
    cerr<<"input is either a single sequence with the -s option or\n";
    cerr<<"a file with a list of sequences with the -i option.\n";
    cerr<<"Standard yeast sequence files and orf tables are expected.\n";
    exit(0);
  }
  string s;
  GetArg2(argc, argv,"-s",s);
  string out;
  GetArg2(argc, argv,"-o", out);
  ofstream fout(out.c_str());
  string in;
  GetArg2(argc, argv,"-i", in);
  ifstream fin(in.c_str());
  vector<string> seqset;
  get_fasta_fast("yeast.seq",seqset);
  Genome g(seqset);
  ifstream tab_orf("ORF_table.txt");
  ifstream tab_nonorf("other_features_table.txt");
  g.load_table(tab_orf,true);
  g.load_table(tab_nonorf,false);
  g.add_ends(seqset);
  int i,j,k;
  for(i=0;i<seqset.size();i++){
    seqset[i]=capitalize(seqset[i]);
  }
  if(s.size()>0){
    s=capitalize(s);
    for(i=0;i<seqset.size();i++){
      k=-1;j=0;
      while((k=seqset[i].find(s,j))!=string::npos){
	j=k+1;
	OrfLocale o(i+1,k,g);
	fout<<s<<'\t'<<i<<'\t'<<k<<'\t'<<'F'<<'\n'<<o.desc()<<'\n';
      }
    }
    s=reverse_comp(s);
    for(i=0;i<seqset.size();i++){
      k=-1;j=0;
      while((k=seqset[i].find(s,j))!=string::npos){
	j=k+1;
	OrfLocale o(i+1,k,g);
	fout<<s<<'\t'<<i<<'\t'<<k<<'\t'<<'R'<<'\n'<<o.desc()<<'\n';
      }
    }
  }
  else{
    while(getline(fin,s,'\n')){
      s=capitalize(s);
      for(i=0;i<seqset.size();i++){
	k=-1;j=0;
	while((k=seqset[i].find(s,j))!=string::npos){
	  j=k+1;
	  OrfLocale o(i+1,k,g);
	  fout<<s<<'\t'<<i+1<<'\t'<<k<<'\t'<<'F'<<'\n'<<o.desc()<<'\n';
	}
      }
      s=reverse_comp(s);
      for(i=0;i<seqset.size();i++){
	k=-1;j=0;
	while((k=seqset[i].find(s,j))!=string::npos){
	  j=k+1;
	  OrfLocale o(i+1,k,g);
	  fout<<s<<'\t'<<i+1<<'\t'<<k<<'\t'<<'R'<<'\n'<<o.desc()<<'\n';
	}
      }
    }
  }
}

